package main // autocert-client provides updates to specified key and certificate files

import (
	"bytes"
	"crypto/rsa"
	"crypto/sha256"
	"crypto/tls"
	"crypto/x509"
	"encoding/pem"
	"errors"
	"flag"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"net/url"
	"os"
	"os/user"
	"path/filepath"
	"reflect"
	"strconv"
	"strings"
	"syscall"
	"time"
)

var cfg = DefaultConfig()

// Config represents the various configuration options passed on the command line
type Config struct {
	fqdn string
	auth string

	autocertHost   string
	autocertCAPath string

	keyPath   string
	certPath  string
	chainPath string

	concatChain bool
}

// Cert represents an x509 certificate (and its hash)
type Cert struct {
	raw     []byte            // raw source byte stream
	cert    *x509.Certificate // the above, as *x509.Certificate (or nil if empty)
	hash    []byte            // sha256 hash of the raw data
	version int               // version (optional)
}

// Key represents an RSA private key (and its hash)
type Key struct {
	raw       []byte          // raw source byte stream, may be encrypted
	hash      []byte          // hash of the raw byte stream
	key       *rsa.PrivateKey // unencrypted, as *rsa.PrivateKey (or nil if empty)
	clear     []byte          // unencrypted byte stream
	clearHash []byte          // hash of unencrypted byte stream
}

// Chain represents a chain of one or more intermediate x509 certificates
type Chain struct {
	raw   []byte  // raw source byte stream
	hash  []byte  // hash of the raw byte stream
	certs []*Cert // individual elements as a slice of Cert objects (see above)
}

// DefaultConfig is... our default Config
func DefaultConfig() *Config {
	return &Config{
		concatChain: false,
		//sslVerify:   true,
	}
}

func init() {
	flag.StringVar(&cfg.fqdn, "fqdn", cfg.fqdn,
		"Fully qualified domain name")
	flag.StringVar(&cfg.auth, "auth", cfg.auth,
		"Authentication key")

	flag.StringVar(&cfg.autocertHost, "autocert_host", cfg.autocertHost,
		"Host to contact")
	flag.StringVar(&cfg.autocertCAPath, "autocert_ca_path", cfg.autocertCAPath,
		"Path to CA cert to authenticate (if no root trust)")

	flag.StringVar(&cfg.certPath, "cert_path", cfg.certPath,
		"Path to certificate file")
	flag.StringVar(&cfg.keyPath, "key_path", cfg.keyPath,
		"Path to key file")
	flag.StringVar(&cfg.chainPath, "chain_path", cfg.chainPath,
		"Path to intermediate chain file")

	flag.BoolVar(&cfg.concatChain, "concat_chain", cfg.concatChain,
		"Whether to concatenate the chain and the cert file (chainPath will be ignored if so)")
}

func fqdnInSAN(match string, sans []string) bool {
	// if we got 'wildcard.example.com', check for '*.example.com' instead
	if strings.HasPrefix(match, "wildcard.") {
		match = "*" + match[8:]
	}
	for _, san := range sans {
		if san == match {
			return true
		}
	}
	return false
}

func readCertFromFile(p string) (*Cert, error) {
	raw, err := readFile(p)
	if err != nil {
		// something went wrong reading the file, return no cert and the error
		return nil, err
	}

	// hash the raw data, which could be empty but at least we have a hash
	hash := hashBytes(raw)

	// initialise with a nil cert, which we'll overwrite if we got one
	c := &Cert{raw: raw, cert: nil, hash: hash, version: 0}

	if raw != nil {
		// if we read something
		blocks, err := parseAndDecryptPEMs(raw, "") // certs aren't encrypted
		if err != nil {
			return c, err
		}
		c.cert, err = x509.ParseCertificate(blocks[0].Bytes)
		if err != nil {
			return c, err
		}
		//Don't do this, cos we may include a chain:
		//c.raw = pem.EncodeToMemory(blocks[0])
		//c.hash = hashBytes(c.raw) // and a better hash, too
	}

	return c, nil
}

func updateCertFromServer(client *http.Client, oldCert *Cert, cfg *Config) (*Cert, error) {
	url, err := formObjURL(cfg.autocertHost, cfg.fqdn, "cert")
	if err != nil {
		return nil, err
	}
	raw, version, err := freshenObject(
		client,
		oldCert.raw,
		url,
		cfg.fqdn,
		oldCert.hash,
		cfg.auth,
		oldCert.version,
	)
	if err != nil {
		return nil, err
	}

	hash := hashBytes(raw)
	blocks, err := parseAndDecryptPEMs(raw, "") // certs aren't encrypted
	if err != nil {
		return nil, err
	}
	cert, err := x509.ParseCertificate(blocks[0].Bytes)
	if err != nil {
		return nil, err
	}

	return &Cert{
		raw:     raw,
		cert:    cert,
		hash:    hash,
		version: version,
	}, nil
}

func readChainFromFile(p string, skip int) (*Chain, error) {
	raw, err := readFile(p)
	if err != nil {
		// something went wrong reading the file, return no chain and the error
		return nil, err
	}

	hash := hashBytes(raw)
	c := &Chain{raw: raw, hash: hash}

	if raw != nil {
		// if we read something, return a full Chain
		blocks, err := parseAndDecryptPEMs(raw, "") // chains aren't encrypted
		if err != nil {
			return c, err
		}
		for _, block := range blocks {
			// we may want to skip the first x blocks (e.g. for a file containing cert then chain)
			if skip > 0 {
				skip--
				continue
			}
			// append each intermediate cert to our slice of Certs
			raw = pem.EncodeToMemory(block)
			hash := hashBytes(raw)

			cert, err := x509.ParseCertificate(block.Bytes)
			if err != nil {
				return c, err
			}

			c.certs = append(c.certs, &Cert{
				cert:    cert,
				raw:     raw,
				hash:    hash,
				version: 0,
			})
		}
		c.raw = []byte{} // clear c.raw for later appending
	}
	for _, chaincert := range c.certs {
		c.raw = append(c.raw, chaincert.raw...)
	}

	c.hash = hashBytes(c.raw)

	return c, nil
}

func updateChainFromServer(client *http.Client, oldChain *Chain, cfg *Config, version int) (*Chain, error) {
	url, err := formObjURL(cfg.autocertHost, cfg.fqdn, "chain")
	if err != nil {
		return nil, err
	}
	raw, _, err := freshenObject(client, oldChain.raw, url, cfg.fqdn, oldChain.hash, cfg.auth, version)
	if err != nil {
		return nil, err
	}
	hash := hashBytes(raw)
	blocks, err := parseAndDecryptPEMs(raw, "") // chains aren't encrypted
	if err != nil {
		return nil, err
	}

	var certs []*Cert

	for _, block := range blocks {
		// append each intermediate cert to our slice of Certs
		raw := block.Bytes
		hash := hashBytes(raw)
		cert, err := x509.ParseCertificate(block.Bytes)
		if err != nil {
			return nil, err
		}
		certs = append(certs, &Cert{
			raw:     raw,
			cert:    cert,
			hash:    hash,
			version: 0,
		})
	}
	return &Chain{
		raw:   raw,
		hash:  hash,
		certs: certs,
	}, nil
}

func readKeyFromFile(p string) (*Key, error) {
	raw, err := readFile(p)
	if err != nil {
		// something went wrong reading the file, return no cert and the error
		return nil, err
	}
	hash := hashBytes(raw)

	k := &Key{raw: raw, hash: hash}

	if raw != nil {
		// if we read something, return a full Key
		blocks, err := parseAndDecryptPEMs(raw, cfg.auth)
		if err != nil {
			return k, err
		}
		key, err := blockToKey(blocks[0])
		k.key = key
		if err != nil {
			return k, err
		}
		// create a decrypted byte stream
		k.clear = pem.EncodeToMemory(blocks[0])
		k.clearHash = hashBytes(k.clear)
	}
	return k, nil
}

func updateKeyFromServer(client *http.Client, oldKey *Key, cfg *Config, version int) (*Key, error) {
	url, err := formObjURL(cfg.autocertHost, cfg.fqdn, "key")
	if err != nil {
		return nil, err
	}
	raw, _, err := freshenObject(client, oldKey.raw, url, cfg.fqdn, oldKey.hash, cfg.auth, version)
	if err != nil {
		return nil, err
	}
	hash := hashBytes(raw)
	k := &Key{raw: raw, hash: hash}
	blocks, err := parseAndDecryptPEMs(raw, cfg.auth) // keys may be encrypted with cfg.auth
	if err != nil {
		return k, err
	}
	key, err := blockToKey(blocks[0])
	k.key = key
	if err != nil {
		return k, err
	}

	// create a decrypted byte stream
	k.clear = pem.EncodeToMemory(blocks[0])
	k.clearHash = hashBytes(k.clear)

	return k, nil
}

func readFile(p string) ([]byte, error) {
	// the file may not (yet) exist, in which case we return an empty byte slice
	content, err := ioutil.ReadFile(p)

	if os.IsNotExist(err) {
		// this is actually fine, the file may not exist yet
		return nil, nil
	}
	return content, err
}

func parseAndDecryptPEMs(b []byte, auth string) ([]*pem.Block, error) {
	// do nothing yet
	var ret []*pem.Block
	for {
		// pull the first block of 'b' into 'block', then drop the remainder back into 'b' for next go around
		block, remainder := pem.Decode(b)
		if block == nil {
			return ret, errors.New("Failed to parse PEM")
		}
		if x509.IsEncryptedPEMBlock(block) {
			// the only blocks we expect to be encrypted are private keys
			derBytes, err := x509.DecryptPEMBlock(block, []byte(auth))
			if err != nil {
				return ret, err
			}
			block = &pem.Block{Type: block.Type, Bytes: derBytes}
		}
		ret = append(ret, block)
		if len(remainder) == 0 {
			break
		}
		b = remainder
	}
	return ret, nil
}

func blockToKey(b *pem.Block) (*rsa.PrivateKey, error) {
	// take a *pem.Block and return an *rsa.PrivateKey
	switch b.Type {
	case "RSA PRIVATE KEY":
		// PKCS#1 format
		return x509.ParsePKCS1PrivateKey(b.Bytes)
	case "PRIVATE KEY":
		// PKCS#8 format
		keyInt, err := x509.ParsePKCS8PrivateKey(b.Bytes)
		if err != nil {
			return nil, err
		}
		key, ok := keyInt.(*rsa.PrivateKey)
		if ok {
			return key, nil
		}
		return key, errors.New("Couldn't parse PKCS#8 key")
	}
	return nil, fmt.Errorf("Couldn't parse key of type: %s)", b.Type)
}

func hashBytes(b []byte) []byte {
	h := sha256.New()
	h.Write(b)
	return h.Sum(nil)
}

func formObjURL(autocertHost string, fqdn string, objType string) (string, error) {
	scheme := "https"
	path := fqdn

	switch objType {
	case "cert":
		path = path + ".crt"
	case "key":
		path = path + ".key"
	case "chain":
		path = path + "_chain.crt"
	default:
		return "", fmt.Errorf("Couldn't form a url for object type %s", objType)
	}

	url := &url.URL{
		Scheme: scheme,
		Host:   autocertHost,
		Path:   path,
	}

	return url.String(), nil
}

func freshenObject(client *http.Client, oldObj []byte, url string, user string,
	hash []byte, pass string, version int) ([]byte, int, error) {

	var obj []byte
	ver := 0

	// define our request
	req, err := http.NewRequest("GET", url, nil)
	if err != nil {
		// if our request couldn't be formed
		return obj, ver, err
	}
	req.SetBasicAuth(user, pass)

	if hash != nil {
		// send our Hash as a header, a dumb static site will ignore this
		req.Header.Set("X-AutoCert-Hash", fmt.Sprintf("%x", hash))
	}
	// send our requested Version (default 0), again a dumb site will ignore this
	req.Header.Set("X-AutoCert-Version", fmt.Sprintf("%d", version))

	// get the response
	resp, err := client.Do(req)

	if err != nil {
		// if the response came with an error
		return obj, ver, err
	}

	defer resp.Body.Close()

	if resp.StatusCode == http.StatusNotModified {
		obj = oldObj // nothing changed, send back the old one
	} else if resp.StatusCode == http.StatusOK {
		// get the version header, if we received one
		verStr := resp.Header.Get("X-AutoCert-Version")
		// parse that version, if provided, as an int
		if verStr != "" {
			ver, err = strconv.Atoi(verStr)
			if err != nil {
				return obj, ver, err
			}
		}
		// pull the actual body (i.e. our object)
		obj, err = ioutil.ReadAll(resp.Body)
		if err != nil {
			// if we failed to read the returned body
			return obj, ver, err
		}
	} else {
		e := fmt.Errorf("Object retrieval failed, server returned status code %d", resp.StatusCode)
		return obj, ver, e
	}
	return obj, ver, nil
}

func writeFileAtomic(path string, content []byte) error {
	myself, err := user.Current()
	if err != nil {
		return fmt.Errorf("unable to lookup current user")
	}
	int64Uid, err := strconv.ParseInt(myself.Uid, 10, 32)
	if err != nil {
		return fmt.Errorf("unable to parse %s as an integer", myself.Uid)
	}
	newFileUID := int(int64Uid)
	int64Gid, err := strconv.ParseInt(myself.Gid, 10, 32)
	if err != nil {
		return fmt.Errorf("unable to parse %s as an integer", myself.Gid)
	}
	newFileGID := int(int64Gid)

	newFilePerms := os.FileMode(0600)

	if fi, err := os.Stat(path); os.IsNotExist(err) {
		// path does not exist, stick with the defaults
	} else {
		newFileUID = int(fi.Sys().(*syscall.Stat_t).Uid)
		newFileGID = int(fi.Sys().(*syscall.Stat_t).Gid)
		newFilePerms = fi.Mode()

		timeStamp := time.Now().Format(time.RFC3339)

		backupPath := fmt.Sprintf("%s.backup.autocert.%s", path, timeStamp)
		if err := os.Rename(path, backupPath); err != nil {
			return fmt.Errorf("unable to backup %s to %s: %v", path, backupPath, err)
		}
	}

	tf, err := ioutil.TempFile(filepath.Dir(path), "autocert")
	if err != nil {
		return fmt.Errorf("unable to create temporary file: %v", err)
	}
	defer os.Remove(tf.Name())
	if _, err := tf.Write(content); err != nil {
		return fmt.Errorf("unable to write temporary file: %v", err)
	}
	if err := tf.Close(); err != nil {
		return fmt.Errorf("unable to close temporary file: %v", err)
	}

	if err := os.Rename(tf.Name(), path); err != nil {
		return fmt.Errorf("unable to move temporary file: %v", err)
	}

	// Apply the old file's ownership to the newly written version
	if err := os.Chown(path, int(newFileUID), int(newFileGID)); err != nil {
		return fmt.Errorf("unable to chown %s with %d:%d", path, newFileUID, newFileGID)
	}

	// Apply the old file's permissions to the newly written version
	if err := os.Chmod(path, newFilePerms); err != nil {
		return fmt.Errorf("unable to chmod %s with %v", path, newFilePerms)
	}

	return nil
}

func main() {
	flag.Parse()

	exitCode := 2 // 2 = nothing happened, 1 = something went wrong, 0 = we freshened stuff, go restart the service

	// we need to know what we're renewing, with what token, and where
	if cfg.fqdn == "" {
		log.Fatal("Required argument fqdn not provided")
	}
	if cfg.auth == "" {
		log.Fatal("Required argument auth not provided")
	}
	if cfg.autocertHost == "" {
		log.Fatal("Required argument autocert_host not provided")
	}

	// these two need to be defined but don't necessarily need to exist
	// an empty chain is acceptable, for now
	if cfg.certPath == "" {
		log.Fatal("Required argument certPath not provided")
	} else if cfg.keyPath == "" {
		log.Fatal("Required argument keyPath not provided")
	}

	// set up our cert pool for service certificate trust
	certPool, err := x509.SystemCertPool()
	if err != nil {
		log.Fatal(err)
	}

	// if autocertCAPath is specified, load and append that cert to our pool
	if cfg.autocertCAPath != "" {
		caCert, err := readCertFromFile(cfg.autocertCAPath)
		if err != nil {
			log.Fatal(err)
		}
		certPool.AddCert(caCert.cert)
	}

	tr := &http.Transport{
		TLSClientConfig: &tls.Config{RootCAs: certPool},
	}

	client := &http.Client{Transport: tr}

	// load the existing cert, if it exists
	oldCert, err := readCertFromFile(cfg.certPath)
	if err != nil {
		log.Fatal(err)
	}
	// and now pull the latest cert from the server (could be new, could be the same)
	newCert, err := updateCertFromServer(client, oldCert, cfg)
	if err != nil {
		log.Fatal(err)
	}

	if newCert.cert == nil {
		log.Fatal("Unhandled error retrieving the cert (content is nil)")
	}

	if newCert.version != 0 {
		log.Printf("INFO: We received version %d of the requested cert\n", newCert.version)
	}

	if oldCert.cert != nil {
		if newCert.cert.NotAfter.Before(oldCert.cert.NotAfter) {
			log.Printf(
				"WARN: New cert expires before old cert (%v vs %v)\n",
				newCert.cert.NotAfter,
				oldCert.cert.NotAfter,
			)
		}
	}

	if newCert.cert.NotAfter.Before(time.Now()) {
		log.Fatalf("New cert already expired on %v\n", newCert.cert.NotAfter)
	}

	if !fqdnInSAN(cfg.fqdn, newCert.cert.DNSNames) {
		log.Fatalf("New certificate is not valid for %s", cfg.fqdn)
	}

	// load the old key, if it exists
	oldKey, err := readKeyFromFile(cfg.keyPath)
	if err != nil {
		log.Fatal(err)
	}
	// and now pull the latest key from the server (could be new, could be the same)
	newKey, err := updateKeyFromServer(client, oldKey, cfg, newCert.version)
	if err != nil {
		log.Fatal(err)
	}

	if newKey.key == nil {
		log.Fatal("Unhandled error retrieving the key (content is nil)")
	}

	// modulus of key matches modulus of cert?
	if newCert.cert.PublicKey.(*rsa.PublicKey).N.Cmp(newKey.key.PublicKey.N) != 0 {
		log.Fatal("New cert/key pair have mis-matched moduli")
	}

	// declare old/new Chain vars, as contents will be conditional
	var oldChain *Chain
	var newChain *Chain

	if cfg.chainPath != "" {
		// if we asked to save chain in its own file
		// load the old chain from disk, if it exists
		oldChain, err = readChainFromFile(cfg.chainPath, 0)
		if err != nil {
			log.Fatal(err)
		}
		newChain, err = updateChainFromServer(client, oldChain, cfg, newCert.version)
		if err != nil {
			log.Fatal(err)
		}
	} else if cfg.concatChain == true {
		// if we asked to save the chain appended to the cert
		// load the old chain from disk, if it exists (skip 1st object as that's the cert)
		oldChain, err = readChainFromFile(cfg.certPath, 1)
		if err != nil {
			log.Fatal(err)
		}
		newChain, err = updateChainFromServer(client, oldChain, cfg, newCert.version)
		if err != nil {
			log.Fatal(err)
		}
	} else {
		log.Println("WARN: No chainPath specified and concatChain is false, so no chain retrieved or written")
	}

	concatChainCertWritten := false

	if newChain != nil {
		for _, cert := range newChain.certs {
			if cert.cert.NotAfter.Before(time.Now()) {
				log.Fatalf("ERR: Intermediate cert (%s) has already expired!\n", cert.cert.Subject.CommonName)
			}
		}
		if reflect.DeepEqual(oldChain.hash, newChain.hash) == false {
			if cfg.concatChain == true {
				log.Printf("INFO: Writing cert and chain to %s\n", cfg.certPath)
				err = writeFileAtomic(cfg.certPath, append(newCert.raw, newChain.raw...))
				if err != nil {
					log.Fatal(err)
				}
				concatChainCertWritten = true
				exitCode = 0
			} else if cfg.chainPath != "" {
				log.Printf("INFO: Writing chain to %s\n", cfg.chainPath)
				err = writeFileAtomic(cfg.chainPath, newChain.raw)
				if err != nil {
					log.Fatal(err)
				}
				exitCode = 0
			}
		} else {
			//log.Printf("INFO: No changes to chain")
		}
	}
	if reflect.DeepEqual(newCert.hash, oldCert.hash) == false && concatChainCertWritten == false {
		if cfg.concatChain == true {
			// new/old cert hashes may simply mismatch because of concatenation, look harder
			if reflect.DeepEqual(newCert.cert, oldCert.cert) == false {
				log.Printf("INFO: Writing cert and chain to %s\n", cfg.certPath)
				err = writeFileAtomic(cfg.certPath, append(newCert.raw, newChain.raw...))
				if err != nil {
					log.Fatal(err)
				}
				exitCode = 0
			}
		} else {
			log.Printf("INFO: Writing certificate to %s\n", cfg.certPath)
			err = writeFileAtomic(cfg.certPath, newCert.raw)
			if err != nil {
				log.Fatal(err)
			}
			exitCode = 0
		}
	} else {
		//log.Printf("INFO: No changes to %s\n", cfg.certPath)
	}

	if !bytes.Equal(newKey.clearHash, oldKey.hash) {
		log.Printf("INFO: Writing key to %s\n", cfg.keyPath)
		// write unencrypted key to disk, we don't support encryption here
		err = writeFileAtomic(cfg.keyPath, newKey.clear)
		if err != nil {
			log.Fatal(err)
		}
		exitCode = 0
	} else {
		//log.Printf("INFO: No changes to %s\n", cfg.keyPath)
	}

	os.Exit(exitCode)
}
