package main

import (
	"encoding/hex"
	"testing"
)

func TestHashBytes(t *testing.T) {
	tests := []struct {
		b    string
		want string
	}{
		{"", "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"},
	}
	for _, test := range tests {
		if got := hex.EncodeToString(hashBytes([]byte(test.b))); got != test.want {
			t.Errorf("hashBytes failed for %v, %v", got, test.want)
		}
	}
}

func TestReadCertFromFile(t *testing.T) {
	tests := []struct {
		label   string
		path    string
		wantVer int
	}{
		{"test1.example.com", "tests/test1.example.com.crt", 3},
		{"test2.example.com", "tests/test2.example.com.crt", 3},
		{"test3.example.com", "tests/test3.example.com.crt", 3},
		{"test4.example.com", "tests/test4.example.com.crt", 3},
		{"test5.example.com", "tests/test5.example.com.crt", 3},
		{"404.invalid", "tests/404.invalid.crt", -1}, // we don't expect to read this
	}
	for _, test := range tests {
		res, err := readCertFromFile(test.path)
		if err != nil {
			t.Errorf("readCertFromFile(%s) failed: %v", test.path, err)
		}
		switch {
		case res.cert == nil && test.wantVer != -1:
			t.Errorf("readCertFromFile(%s) returned empty cert", test.path)
		case res.cert != nil && test.wantVer != res.cert.Version:
			t.Errorf("readCertFromFile(%s) version = %d, want %d", test.path, res.cert.Version, test.wantVer)
		}
	}
}

func TestReadKeyFromFile(t *testing.T) {
	tests := []struct {
		label           string
		path            string
		wantPubExponent int
	}{
		{"test1.example.com", "tests/test1.example.com.key", 65537},
		{"test2.example.com", "tests/test2.example.com.key", 65537},
		{"test3.example.com", "tests/test3.example.com.key", 65537},
		{"test4.example.com", "tests/test4.example.com.key", 65537},
		{"test5.example.com", "tests/test5.example.com.key", 65537},
		{"404.invalid", "tests/404.invalid.key", -1},
	}
	for _, test := range tests {
		got, err := readKeyFromFile(test.path)
		if err != nil {
			t.Errorf("readKeyFromFile(%s) failed: %v", test.path, err)
		}
		switch {
		case got.key == nil && test.wantPubExponent != -1:
			t.Errorf("readKeyFromFile(%s) returned empty key", test.path)
		case got.key != nil && test.wantPubExponent != got.key.PublicKey.E:
			t.Errorf(
				"readKeyFromFile(%s) PubKey exponent = %d, want %d",
				test.path,
				got.key.PublicKey.E,
				test.wantPubExponent,
			)
		}
	}
}

func TestFQDNInSAN(t *testing.T) {
	tests := []struct {
		fqdn     string
		certPath string
		want     bool
	}{
		{"test1.example.com", "tests/test1.example.com.crt", true},
		{"www.test1.example.com", "tests/test1.example.com.crt", true},
		{"test1.example.com", "tests/test2.example.com.crt", false},
		{"test2.example.com", "tests/test2.example.com.crt", true},
	}
	for _, test := range tests {
		cert, err := readCertFromFile(test.certPath)
		if err != nil {
			t.Errorf("readCertFromFile(%s) failed: %v", test.certPath, err)
		}
		got := fqdnInSAN(test.fqdn, cert.cert.DNSNames)
		if got != test.want {
			t.Errorf("fqdnInSAN(%s) failed", test.fqdn)
		}
	}
}
